unit timerunit;

{ apply CPU clock counter

  - set CPU clock first by calling "setCPUclock"
  - get CPU clock count by calling "getCPUticks(t)"  twice for t1, t2
  - calculate time (usecs) in usecs by calling "getTime(t2-t1)"

}

interface

uses windows;

procedure setCPUclock;
procedure getCPUticks(var count : Int64);
function ProcTime(t : Int64) : double;

var CPUclock : double;    //CPU clock speed in MHz

implementation

procedure getCPUticks(var count: Int64);
//store 64 bits CPU clock in variable count
begin
 asm
  mov ECX,count;
  RDTSC;          //lower 32 bits --> EAX, upper 32 bits ---> EDX
                  //RDTSC = DB $0F,$31
  mov [ECX],EAX;
  add ECX,4;
  mov [ECX],EDX;
 end;
end;

procedure setCPUclock;
//set variable CPUclock
//call this procedure before measuring any process time
var t1,t2 : cardinal;         //system clock ticks
    cput1,cput2 : Int64;      //CPU clock ticks
begin
 t1 := getTickCount;          //get milliseconds clock
 while getTickCount = t1 do;  //sync with start of 1 millisec interval
 getCPUticks(cput1);          //get CPU clock count
 t1 := gettickcount;
 repeat
  getCPUticks(cput2);         //get CPU clock count
  t2 := gettickcount;
 until t2-t1 >= 500;
 CPUclock := (cput2-cput1)/((t2-t1)*1e3);      //set CPU clock in microsecs
end;

function ProcTime(t : Int64) : double;
//return clock time in usecs
begin
 result := t / CPUclock;
end;

end.
