unit xfontUnit;
{
    ------------------
   | DavData software |
   | rev. 1           |
   | dec. 2011        |
    ------------------

   status:
   3 - 12 - 2011 : fixed cliprect problems 

  the Xfont Class  use home brew scalable fonts, supplied in Timage components
  ----------------

  properties       - textmap : TXbitmap           //w   ...bitmap for write
                   - fontNr : byte;               //w,r ...font type selection
                   - maxfont : byte;              //r   ...highest fontNr
                   - fontName[n] : string;        //r   ...retrieve font name
                   - height byte;                 //w,r ...char size
                   - style : Txfontstyle;         //w,r ...char style
                   - pencolor : DWORD;            //w,r ...char color
                   - brushcolor : DWOrd;          //w,r ...background color
                   - penlevel : byte;             //w,r ...highest 0 lowest 3
                   - brushlevel : byte;           //w,r
                   - fontProps : TXfontProps;     //r   ...font properties
                   - charmetrics : TXcharMetrics; //r   ...char layout
                   - textrect : Trect;            //r,w ...drawing boundaries
                   - modtrect : Trect;            //r   ...modified area
                   - maxfontNr : byte;            //r   ...nr of inst. fonts 1..
                   - posX : smallInt;             //w,r ...x position of char
                   - posY : smallInt;             //w,r ...y position of char

  methods          - create;
                   - destroy;
                   - installfont(im:Timage) : byte;//from bitmap in Timage
                   - charwidth(ch : byte);         //get width of char
                   - charOut(ch : byte);           //write ch at (posX,posY)
}

interface

uses windows,extctrls,Xbitmap;

const maxfontcount  = 3;  //max number of installable fonts
      xfBold       = $1;
      xfItalic     = $2;
      xfWide       = $4;
      xfBrush      = $8;
      xfUnderline  = $10;

type TXfontdata = array[0..20000] of DWORD;  //array with stripe codes....
     PXFont     = ^TXfontData;
     TXFontProps = record
                    name       : string;
                    length     : word;       //length of fonttable DWORDS
                    resolution : byte;
                    driver     : byte;       //1 : stripe driver
                    revision   : byte;
                    baseline   : byte;
                    HiTop      : byte;       //top of 'A'
                    LoTop      : byte;       //top of 'a'
                    PXfont     : PXFont;     //points to fontlist
                   end;

     TcharMetrics = record
                     code : byte;
                     width : byte;
                     hspace : byte;
                     top : byte;
                     bottom : byte;
                    end;

     TPart   = record
                 LT   : byte;
                 bias : smallInt;
                 mult : single;
                end;
     Tpartition = array[1..5] of Tpart;

    TXfont = class
     private
      FTextmap     : TXbitmap;  //may be Tbitmap also
      FfontNr      : byte;      //selected font number
      FmaxFont     : byte;      //highest installed font
      Ffontlist    : array[1..maxfontcount] of TXFontProps;//installed fonts
      Fstyle       : byte;
      Fpencolor    : DWORD;
      FBrushcolor  : DWORD;
      Fpenlevel    : byte;
      Fbrushlevel  : byte;
      Fcliprect    : Trect;
      Fmodrect     : Trect;
      FposX        : smallInt; //signed 16 bit
      FposY        : smallInt;
      Fwidth       : byte;
      Fheight      : byte;

      Flinebase    : DWORD;    //text canvas pointer to (0,0)
      FLineStep    : DWORD;
      FGred        : byte;
      FGgreen      : byte;
      FGblue       : byte;
      YFmult       : single;
      XFmult       : single;
      XFboldexpansion : byte;
      HP           : byte;      //active partition indexes
      VP           : byte;      //..
      ITtopshift   : byte;      //italic : right shift of left top
      ITdx         : single;    //italic : left shift step per y upwards
      Hpartition   : Tpartition;
      VPartition   : Tpartition;
      px1          : byte;      //pixel coordinates in char image
      px2          : byte;      //..
      py           : byte;      //..

     protected
      procedure setfontNr(nr : byte);
      procedure settextMap(bm : TXBitmap);
      procedure unpackSlides(var part:Tpartition; dw:DWORD; f:single);
      procedure paintBackGround;
      procedure paintpixels;
      procedure paintUnderline;
      procedure markstripe(x1,x2,y : byte);
      procedure makeFwidth(w : byte);
      procedure makeModrect;
      procedure setFontHeight(h : byte);
      function getstripe1Ptr(ch : byte) : PDW;
      procedure setFontStyle(st : byte);
      procedure setFGColor(clr : DWORD);
      procedure setBGcolor(clr : DWORD);
      function getFGcolor : DWORD;
      function getBGcolor : DWORD;
      procedure setpenlevel(lvl : byte);
      procedure setbrushlevel(lvl : byte);

     public
      constructor create;
      function install(im : Timage) : byte;//result = fontNr <> 0 if OK
      procedure charOut(ch : byte);
      function fontprops : TXfontprops;
      function charmetrics(ch : byte) : Tcharmetrics;
      function charWidth(ch : byte) : byte;
      property textmap : TXbitmap write setTextmap;
      property posX : smallint read FposX write FposX;
      property posY : smallint read FposY write FposY;
      property maxfont : byte read FmaxFont;
      property FontNr : byte read FFontNr write setfontNr;
      property cliprect : Trect read Fcliprect write Fcliprect;
      property modrect : Trect read Fmodrect;
      property FontHeight : byte read FHeight write setFontHeight;
      property FontStyle : byte read FStyle write setFontStyle;
      property penColor : DWORD read getFGcolor write setFGcolor;
      property brushColor : DWORD read getBGcolor write setBGcolor;
      property PenLevel : byte read FPenlevel write setpenlevel;
      property Brushlevel : byte read FBrushlevel write setBrushlevel;

     end;

implementation

const ITtangent = 0.25;    //italic slope of vertical left 
      boldExpansion = 4;   //bold width increase in pixels
      wideFactor = 1.33;   //wide hor. multiplication factor

constructor TXfont.create;
begin
 Ftextmap := nil;
 FfontNr := 0;
 FmaxFont := 0;
 FontHeight := 20;
 FontStyle := 0;
 Pencolor := $000000;
 BrushColor := $ffffff;
 penlevel := 0;
 brushlevel := 1;
 posX := 0;
 posY := 0;
end;

function TXfont.install(im : Timage) : byte;
//font in bitmap of Timage component
//if maximal font installed: return 0
//else add font, return number 1,2......
var j,b : byte;
    d : DWORD;
    s : string;
begin
 if Fmaxfont = maxfontcount then
  begin
   result := 0;
   exit;
  end;
 inc(FmaxFont);
 FFontNr := FmaxFont;
 result := FFontNr;

 with FFontList[FFontNr] do         //PXfont is start of font data
  begin
   with im.picture.bitmap do PXfont := scanline[height-1];
    d := PXfont^[0];
    driver := d shr 24;
    revision := (d shr 16) and $ff;
    length := d and $ffff;
    s := '';
    d:= PXfont^[1];

    for j := 0 to 3 do
     begin
      b := (d shr (j*8)) and $ff;
      if b <> 0 then s := s + chr(b);
     end;

    d := PXfont^[2];
    for j := 0 to 3 do
     begin
      b := (d shr (j*8)) and $ff;
      if b <> 0 then s := s + chr(b);
     end;
    name := s;

    d := PXfont^[3];
    resolution := d shr 24;
    baseline := d and $ff;
    LoTop  := (d shr 8) and $ff;
    HiTop := (d shr 16) and $ff;
   end;
end;

procedure TXfont.setTextMap(bm : TXBitmap);
//select XBitmap for text output
//set cliprect to total surface
//set base address and line decrement
begin
 FTextMap := bm;
 Flinebase := DWORD(bm.scanline[0]);
 Flinestep := FlineBase - DWORD(bm.scanline[1]);
 with FclipRect do
  begin
   left := 0; right := bm.width;
   top := 0; bottom := bm.height;
  end;
end;

procedure TXfont.setfontNr(nr : byte);
begin
 if (nr> 0) and (nr <= maxfont) then FFontNr := nr;
end;

procedure TXFont.setFontStyle(st : byte);
begin
 Fstyle := st;
 if Fstyle and xfbold <> 0 then xfboldexpansion := boldexpansion
  else xfboldexpansion := 0;
end;

procedure TXFont.setFontHeight(h : byte);
//adjust scale factors,expansion acc to style
begin
 Fheight := h;
 YFmult := h/FFontList[FFontNr].resolution;
 ITdx := ITtangent * YFmult;
 ITtopshift := trunc(FFontList[FFontNr].baseline * ITdx + 0.5);
end;

procedure TXFont.setFGColor(clr : DWORD);//blue bits 01: level
//Fpencolor in format rrggbb
const mask = $ff;
begin
 Fpencolor := swapcolor(clr);
 FPenlevel := FPencolor and mask;
 FGred := (FPencolor shr 16) and mask;
 FGgreen := (FPencolor shr 8) and mask;
 FGblue := Fpencolor and mask;
end;

procedure TXFont.setBGcolor(clr : DWORD);
//Fbrushcolor in format rrggbb
begin
 FBrushcolor := swapcolor(clr);
 FBrushlevel := FBrushcolor and $3;
end;

function TXFont.getFGcolor : DWORD;
begin
 result:= swapcolor(Fpencolor);
end;

function TXFont.getBGcolor : DWORD;
begin
 result := swapcolor(FBrushColor);
end;

procedure TXFont.setpenlevel(lvl : byte);
begin
 lvl := lvl and $3;
 FPenlevel := lvl;
 FPencolor := Fpencolor and $00fffffc or lvl;
 FGBlue := FGBlue and $fc or lvl;
end;

procedure TXFont.setbrushlevel(lvl : byte);
begin
 lvl := lvl and $3;
 FBrushlevel := lvl;
 FPencolor := Fpencolor and $00fffffc or lvl;
end;

function TXfont.Fontprops : TXfontprops;
//show properties of font Nr
begin
 with FFontList[FFontNr] do
  begin
   result.name       := name;
   result.length     := length;       //length of fonttable DWORDS
   result.resolution := resolution;
   result.driver     := driver;       //1 : stripe driver
   result.revision   := revision;
   result.baseline   := baseline;
   result.HiTop      := HiTop;       //top of 'A'
   result.LoTop      := LoTop;       //top of 'a'
   result.PXfont     := PXFont;      //points to fontlist
  end;
end;

function TXfont.CharWidth(ch : byte) : byte;
//return character width acc to height and style
var p : PDW;
    w : byte;
begin
 p := getstripe1Ptr(ch);         //start of stripe list
 w := (p^ shr 8) and $ff;        //extract width
 makeFwidth(w);
 result := FWidth;
end;

function TXfont.CharMetrics(ch : byte) : Tcharmetrics;
//supply char layout in 96 pixel resolution
const bytemask = $ff;
var p : PDW;
begin
 p := getstripe1Ptr(ch);
 with result do
  begin
   code := (p^ shr 16) and bytemask;//if code = 0, ch is undefined
   width  := (p^ shr 8) and bytemask;
   inc(p);
   hspace := p^ and bytemask;
   top := (p^ shr 8) and bytemask;
   bottom := (p^shr 16) and bytemask;
  end;
end;

procedure TXfont.paintBackGround;
//paint background in XBGColor
var i,j : integer;
    p0,p1 : DWORD;
begin
 for j := FposY to FposY + Fheight-1 do
  begin
   p0 := FlineBase - j * FLinestep;
   for i := FPosX to FPosX + Fwidth - 1 do
    begin
     p1 := p0 + 4*i;
     if PDW(p1)^ and $3 >= FBrushlevel then PDW(p1)^ := FBrushColor;
    end;
  end;
end;

procedure TXFont.unpackSlides(var part : Tpartition; dw : DWORD; f : single);
//unpack slides from dw, make partitions in part
type TXslide = record
                 pos96 : byte;   //in 96 * 96
                 Fpos  : single; //in chracter size
                 Rpos  : byte;   //--rounded
               end;
var i,s : byte;
    slides : array[1..6] of TXslide;
    count : byte;
begin
 i := 1;
  with slides[i] do
   begin
    pos96 := 0;                  //slide1 = left side of char
    fpos := 0;
    rpos := 0;
   end;
 while dw <> 0 do                //copy slides from stripe(font)list
  begin
   inc(i);
   with slides[i] do
    begin
     s := dw and $7f;
     pos96 := s;                 //relative to resolution
     fpos := s*f;                //relative to character size
     rpos := trunc(fpos + 0.5);  //rounded to nearest integer
     dw := dw shr 7;
    end;
  end;//while
 inc(i);
 with slides[i] do               //right side of character
  begin
   pos96 := 96;
   fpos := 96*f;
   rpos := trunc(fpos + 0.5);
  end;
 count := i-1;

//---- here room for extra processing of slide rounding ----

 for i := 1 to count do
  with part[i] do
   begin
    LT := slides[i].pos96;         //LeftTop
    bias := slides[i].rpos;
    mult := (slides[i+1].rpos-bias)/(slides[i+1].pos96-LT);
   end;
end;

procedure TXfont.paintpixels;
//scan pixels and fill colors
//from (FposX,FposY)
const pcor = 0.004; //250 --> 1 correction
var j,r,g,b,pc,pc1,level : byte;
    p0,p1 : DWORD;
    BG : DWORD;
    i,x1,x2,y : integer;
begin
 if Fwidth = 0 then exit;

 for j := 0 to Fheight-1 do
  begin
   y := FPosY+j;
   with FCliprect do if (y < top) or (y >= bottom) then continue;//

   p0 := Flinebase - Flinestep*y;

   if Fstyle and XFitalic <> 0 then
    begin
     i := trunc(FPosX + ITtopshift - ITtangent*j);
     x1 := trunc(i);
     x2 := trunc(i + Fwidth);
    end
    else
     begin
      x1 := FPosX;
      x2 := FposX + Fwidth;
     end;

    with FCliprect do                              //clip horizontal
     begin
      if (x1 >= Right) or (x2 < left) then continue;
      if x1 < left then x1 := left;
      if x2 >= right then x2 := right-1;
     end;

   for i := x1 to x2 do
    begin
     p1 := p0 + 4*i;
     BG := PDW(p1)^;
     pc := BG shr 24;
     BG := BG and $00ffffff;
     if (pc <> 0) and (penlevel <= BG and $3) then
      begin
       pc1 := 250 - pc;
       b := BG and $ff;
       g := (BG shr 8) and $ff;
       r := (BG shr 16) and $ff;
       b := trunc((b*pc1 + FGblue*pc)*pcor) and $fc or FPenlevel;
       g := trunc((g*pc1 + FGgreen*pc)*pcor);
       r := trunc((r*pc1 + FGred*pc)*pcor);
       BG := b or (g shl 8) or (r shl 16);
      end; //if pc
     PDW(p1)^ := BG;
    end;//for i
  end;//for j
end;

procedure TXFont.markstripe(x1,x2,y : byte);
//x1,x2,y : relative position in character
//FPosX,FPosY : screen coordinates
//HP,VP for partitions
//Fstyle (normal,wide,italic,bold,underlined)
//write fill% in byte 3 of pixel
var shiftX,FX1,FX2,FY1,FY2,dx,dy : single;
    p0,p1 : DWORD;
    i,j : integer;
    sum : word;
begin

//-- adjust x1,x2 acc. to fontstyle --

 x2 := x2 + XFboldExpansion;

 if Fstyle and xfItalic <> 0 then ShiftX := ITtopshift - y * ITdx
  else ShiftX := 0;

//-- calculate stripe --

 with Hpartition[HP] do
  begin
   FX1 := FposX + Bias + (x1-LT) * mult + shiftX;
   FX2 := FPosX + Bias + (x2-LT) * mult + ShiftX;
  end;
 with VPartition[VP] do
  begin
   FY1 := FPosY + Bias + (y-LT) * mult;
   FY2 := FY1 + mult;
  end;

 with Fcliprect do                //clip 
  begin
   if (FY1 < top) or (FY2 >= bottom) then exit;
   if (FX1 >= right) or (FX2 < left) then exit;
   if (FX1 < left) then FX1 := left;
   if (FX2 >= right) then FX2 := right-1;
  end;

//-- update intensities in byte 3 of pixels

 for j := trunc(FY1) to trunc(FY2) do
  begin
   if j < FY1 then dy := j+1-FY1 else dy := 1;
   if j+1 > FY2 then dy := dy - (j+1-FY2);
   p0 := Flinebase - j*FlineStep + 3;
   for i := trunc(FX1) to trunc(FX2) do  
    begin
     if i < FX1 then dx := i+1-FX1 else dx := 1;
     if i+1 > FX2 then dx := dx - (i+1-FX2);
     p1 := p0 + 4*i;
     sum := PByte(p1)^ + trunc(dx*dy*250);
     if sum > 250 then sum := 250;//bold may have overlapping stripes
     PByte(p1)^ := sum;
    end;//for i
  end;//for j
//--
end;

procedure TXFont.paintUnderline;
var d,i : byte;
    y : smallInt;
begin
 d := FHeight div 24 + 1;
 y := FposY + FHeight - d;
 with Ftextmap do
  begin
   XpenColor := FPenColor;
   XpenWidth := 1;
   XPenlevel := FPenlevel;
   Xlinestyle := 0;
   XarrowCode := 0;
   for i := 0 to d-1 do
   Xline(FposX,y+i,FposX+Fwidth-1,y+i);
  end;
end;

procedure TXFont.makeFwidth(w : byte);
//w is width for 96*96 image
//set Fwidth acc. to style and height
//this is the central place to calculate the width of a character
//and set XFmult
begin
 if FStyle and XFbold <> 0 then w := w + boldexpansion;
 if FStyle and XFwide <> 0 then w := trunc(w * widefactor + 0.5);
 FWidth := trunc(w * YFmult + 0.5);
 XFmult := FWidth/w;
end;

procedure TXfont.makeModrect;
//area of textmap that was modified at last charout()
begin
with Fmodrect do
 begin
  left := FPosX;
  top := FPosY;
  right := FPosX + Fwidth + 1;
  if FStyle and XFitalic <> 0 then right := right + itTopShift;
  bottom := FPosY + FHeight + 1;
 end; 
end;

function TXFont.getstripe1Ptr(ch : byte) : PDW;
//get first location of stripe list for character ch
var dw,ptr : dword;
begin
 ptr := DWORD(FFontList[FFontNr].PXfont);//points to top of fontlist
 dw := ((ch shr 1) + 4) shl 2;            //byte index to slide address
 dw := PDW(dw + ptr)^;
 if (ch and 1) = 1 then dw := dw shr 16 else dw := dw and $ffff;
 if dw = 0 then dw := 132;//if char not defined, display chr(0)
 dw := (dw shl 2) + ptr;
 result := PDW(dw);
end;

procedure TXFont.charOut(ch : byte);
//write ch at (posX,posY) on XfontMap
//stripelist starts at address p
//write within XFcliprect
const sh : array[0..2] of byte = (0,10,20); //stripe position in list
var pp,maxpp : DWORD;
    dc,i,rep : byte;
    stripe : word;
    delta : shortInt;
    DWstripe : DWORD;
    p : PDW;
begin
 p := getstripe1Ptr(ch);
 HP := 1;  //hor partition #
 VP := 1;  //vert --
 px1 := 0;
 px2 := 0;
 py := 0;
 pp := DWORD(p);           // stripelist #1

 DWstripe := PDW(pp)^;
 maxpp := pp + ((DWstripe and $ff)-1)*4;

 makeFwidth((DWstripe shr 8) and $ff);

 if Fstyle and XFBrush = XFbrush then paintbackground;

 inc(pp,4);                                   // --> stripelist #2
//---
 inc(pp,4);                                   // --> stripelist #3
 unpackSlides(Hpartition,PDW(pp)^,XFmult);
 inc(pp,4);                                   // --> stripelist #4
 unpackslides(VPartition,PDW(pp)^,YFmult);
 inc(pp,4);                                   // --> stripelist #5,6,7,...

 while pp <= maxpp do    //execute stripes list
  begin
   DWstripe := PDW(pp)^;
   for dc := 0 to 2 do
    begin
     stripe := (DWstripe shr sh[dc]) and $3ff;
      case stripe and $380 of
       $000 : py := stripe and $7f;
       $080 : px1 := stripe and $7f;
       $100 : px2 := stripe and $7f;
       $180 : begin
               rep := stripe and $7f;
               for i := 1 to rep do
                begin
                 markstripe(px1,px2,py);
                 inc(py);
                end;
              end;
       $200,
       $280 : begin
               markstripe(px1,px2,py);
               delta := (stripe shr 4) and $f;
               if delta > 7 then delta := delta - 16;
               px1 := px1 + delta;
               delta := stripe and $f;
               if delta > 7 then delta := delta - 16;
               px2 := px2 + delta;
               inc(py);
              end;
       $300,
       $380 : begin
               HP := (stripe shr 4) and $f;
               VP := stripe and $f;
              end;
      end;//case stripe
     end;//for
   inc(pp,4);
  end;//while

 paintpixels;
 if (Fstyle and XFUnderline) <> 0 then paintUnderline;
 makemodrect;
 posX := posX + FWidth;
end;

end.

